@@ -29,13 +29,34 @@
   "The EasyPG Assistant hooks for transparent file encryption"
   :group 'epa)
 
-(defcustom epa-file-name-regexp "\\.gpg\\'"
-  "Regexp which matches filenames to be encrypted with GnuPG."
+(defun epa-file--file-name-regexp-set (variable value)
+  (set-default variable value)
+  (if (fboundp 'epa-file-name-regexp-update)
+      (epa-file-name-regexp-update)))
+
+(defcustom epa-file-name-regexp "\\.gpg\\(~\\|\\.~[0-9]+~\\)?\\'"
+  "Regexp which matches filenames to be encrypted with GnuPG.
+
+If you set this outside Custom while epa-file is already enabled, you
+have to call `epa-file-name-regexp-update' after setting it to
+properly update file-name-handler-alist.  Setting this through Custom
+does that automatically."
   :type 'regexp
-  :group 'epa-file)
+  :group 'epa-file
+  :set 'epa-file--file-name-regexp-set)
 
 (defcustom epa-file-cache-passphrase-for-symmetric-encryption nil
-  "If t, cache passphrase for symmetric encryption."
+  "If non-nil, cache passphrase for symmetric encryption."
+  :type 'boolean
+  :group 'epa-file)
+
+(defcustom epa-file-inhibit-auto-save t
+  "If non-nil, disable auto-saving when opening an encrypted file."
+  :type 'boolean
+  :group 'epa-file)
+
+(defcustom epa-file-select-keys nil
+  "If non-nil, always asks user to select recipients."
   :type 'boolean
   :group 'epa-file)
 
@@ -55,11 +76,29 @@ May either be a string or a list of strings.")
 			val)
 		  t)))))
 
+;;;###autoload
+(put 'epa-file-encrypt-to 'permanent-local t)
+
 (defvar epa-file-handler
   (cons epa-file-name-regexp 'epa-file-handler))
 
 (defvar epa-file-passphrase-alist nil)
 
+(eval-and-compile
+  (if (fboundp 'encode-coding-string)
+      (defalias 'epa-file--encode-coding-string 'encode-coding-string)
+    (defalias 'epa-file--encode-coding-string 'identity)))
+
+(eval-and-compile
+  (if (fboundp 'decode-coding-string)
+      (defalias 'epa-file--decode-coding-string 'decode-coding-string)
+    (defalias 'epa-file--decode-coding-string 'identity)))
+
+(defun epa-file-name-regexp-update ()
+  (interactive)
+  (unless (equal (car epa-file-handler) epa-file-name-regexp)
+    (setcar epa-file-handler epa-file-name-regexp)))
+
 (defun epa-file-passphrase-callback-function (context key-id file)
   (if (and epa-file-cache-passphrase-for-symmetric-encryption
 	   (eq key-id 'SYM))
@@ -104,8 +143,8 @@ May either be a string or a list of strings.")
 	   (point-min) (point-max)
 	   (substring file 0 (string-match epa-file-name-regexp file))
 	   visit beg end replace)))
-    (insert (decode-coding-string string (or coding-system-for-read
-					     'undecided)))))
+    (insert (epa-file--decode-coding-string string (or coding-system-for-read
+						       'undecided)))))
 
 (defvar last-coding-system-used)
 (defun epa-file-insert-file-contents (file &optional visit beg end replace)
@@ -113,15 +152,16 @@ May either be a string or a list of strings.")
   (if (and visit (or beg end))
       (error "Attempt to visit less than an entire file"))
   (setq file (expand-file-name file))
-  (let ((local-copy (epa-file-run-real-handler #'file-local-copy (list file)))
-	(context (epg-make-context))
-	string length entry)
+  (let* ((local-copy (epa-file-run-real-handler #'file-local-copy (list file)))
+	 (local-file (or local-copy file))
+	 (context (epg-make-context))
+	 string length entry)
     (if visit
 	(setq buffer-file-name file))
     (epg-context-set-passphrase-callback
      context
      (cons #'epa-file-passphrase-callback-function
-	   file))
+	   local-file))
     (epg-context-set-progress-callback context
 				       #'epa-progress-callback-function)
     (unwind-protect
@@ -129,12 +169,15 @@ May either be a string or a list of strings.")
 	  (if replace
 	      (goto-char (point-min)))
 	  (condition-case error
-	      (setq string (epg-decrypt-file context file nil))
+	      (setq string (epg-decrypt-file context local-file nil))
 	    (error
 	     (if (setq entry (assoc file epa-file-passphrase-alist))
 		 (setcdr entry nil))
 	     (signal 'file-error
 		     (cons "Opening input file" (cdr error)))))
+	  (make-local-variable 'epa-file-encrypt-to)
+	  (setq epa-file-encrypt-to
+		(mapcar #'car (epg-context-result-for context 'encrypted-to)))
 	  (if (or beg end)
 	      (setq string (substring string (or beg 0) end)))
 	  (save-excursion
@@ -156,36 +199,47 @@ May either be a string or a list of strings.")
       (error "Can't append to the file."))
   (setq file (expand-file-name file))
   (let* ((coding-system (or coding-system-for-write
-			    (if (boundp 'last-coding-system-used)
-				(condition-case nil
-				    (write-region (point-min) (point-max) "/")
-				  (error last-coding-system-used))
+			    (if (fboundp 'select-safe-coding-system)
+				;; This is needed since Emacs 22 has
+				;; no-conversion setting for *.gpg in
+				;; `auto-coding-alist'.
+			        (let ((buffer-file-name
+				       (file-name-sans-extension file)))
+				  (select-safe-coding-system
+				   (point-min) (point-max)))
 			      buffer-file-coding-system)))
 	 (context (epg-make-context))
 	 (coding-system-for-write 'binary)
-	 string entry)
+	 string entry
+	 (recipients
+	  (cond
+	   ((listp epa-file-encrypt-to) epa-file-encrypt-to)
+	   ((stringp epa-file-encrypt-to) (list epa-file-encrypt-to)))))
     (epg-context-set-passphrase-callback
      context
      (cons #'epa-file-passphrase-callback-function
 	   file))
     (epg-context-set-progress-callback context
 				       #'epa-progress-callback-function)
+    (epg-context-set-armor context epa-armor)
     (condition-case error
 	(setq string
 	      (epg-encrypt-string
 	       context
 	       (if (stringp start)
-		   (encode-coding-string start coding-system)
-		 (encode-coding-string (buffer-substring start end)
-				       coding-system))
-	       (unless (assoc file epa-file-passphrase-alist)
-		 (epa-select-keys
-		  context
-		  "Select recipents for encryption.
+		   (epa-file--encode-coding-string start coding-system)
+		 (epa-file--encode-coding-string (buffer-substring start end)
+						 coding-system))
+	       (if (or epa-file-select-keys
+		       (not (local-variable-p 'epa-file-encrypt-to
+					      (current-buffer))))
+		   (epa-select-keys
+		    context
+		    "Select recipents for encryption.
 If no one is selected, symmetric encryption will be performed.  "
-		  (cond
-		   ((listp epa-file-encrypt-to) epa-file-encrypt-to)
-		   ((stringp epa-file-encrypt-to) (list epa-file-encrypt-to)))))))
+		    recipients)
+		 (if epa-file-encrypt-to
+		     (epg-list-keys context recipients)))))
       (error
        (if (setq entry (assoc file epa-file-passphrase-alist))
 	   (setcdr entry nil))
@@ -209,6 +263,23 @@ If no one is selected, symmetric encryption will be performed.  "
 	(message "Wrote %s" buffer-file-name))))
 (put 'write-region 'epa-file 'epa-file-write-region)
 
+(defun epa-file-find-file-hook ()
+  (if (and buffer-file-name
+	   (string-match epa-file-name-regexp buffer-file-name)
+	   epa-file-inhibit-auto-save)
+      (auto-save-mode 0))
+  (set-buffer-modified-p nil))
+
+(defun epa-file-select-keys ()
+  "Select recipients for encryption."
+  (interactive)
+  (make-local-variable 'epa-file-encrypt-to)
+  (setq epa-file-encrypt-to
+	(epa-select-keys
+	 (epg-make-context)
+	 "Select recipents for encryption.
+If no one is selected, symmetric encryption will be performed.  ")))
+
 ;;;###autoload
 (defun epa-file-enable ()
   (interactive)
@@ -216,6 +287,7 @@ If no one is selected, symmetric encryption will be performed.  "
       (message "`epa-file' already enabled")
     (setq file-name-handler-alist
 	  (cons epa-file-handler file-name-handler-alist))
+    (add-hook 'find-file-hooks 'epa-file-find-file-hook)
     (message "`epa-file' enabled")))
 
 ;;;###autoload
@@ -225,6 +297,7 @@ If no one is selected, symmetric encryption will be performed.  "
       (progn
 	(setq file-name-handler-alist
 	      (delq epa-file-handler file-name-handler-alist))
+	(remove-hook 'find-file-hooks 'epa-file-find-file-hook)
 	(message "`epa-file' disabled"))
     (message "`epa-file' already disabled")))
 
